;;; -*- syntax: common-lisp; package: clm; base: 10; mode: lisp -*-
;;;
;;; Envelope handlers
;;;
;;;   eref env x                               value of env at x
;;;   env-length env                           number of break points in env
;;;   env-reverse env                          turn env backwards
;;;   env-concatenate envs                     concatenate envs into one env
;;;   merge-breakpoints brkpts                 sort breakpoints and merge into env
;;;   env-funcall, env-apply, env-map          apply or map some function to or across envelope(s)
;;;   env+ envs...                             add all envs together
;;;   env* envs...                             multiply envs together 
;;;   env-simplify env &optional ygrid xgrid   simplify env
;;;   fft-env-simplify env &optional cutoff    same but use fft filtering
;;;   env-repeat                               repeat/reflect env


(in-package :clm)

#-cltl2 (export '(EREF ENV-LENGTH ENV-REVERSE ENV-CONCATENATE ENV+ ENV* ENV-MAX ENV-REPEAT
		  ENV-FUNCALL ENV-APPLY ENV-MAP MAP-ACROSS-ENVELOPES
		  ADD-OR-EDIT-BREAKPOINT REMOVE-BREAKPOINT ENV-SIMPLIFY FFT-ENV-SIMPLIFY
		  REDUCE-AMPLITUDE-QUANTIZATION-NOISE MELD-ENVELOPES
		  ))

(defun eref (env x) (list-interp x env))

(defun env-length (e) (floor (length e) 2))

(defun env-reverse (e)
  (let ((len (length e)))
    (if (or (= len 0) (= len 2))
	e
      (let ((xmax (nth (- len 2) e))
	    (ne nil))
	(loop for x in e by #'cddr and y in (cdr e) by #'cddr do
	  (push y ne)
	  (push (- xmax x) ne))
	ne))))

(defun merge-breakpoints (&rest breakpoints)
  (if breakpoints
      (apply #'append 
	     (sort (remove-duplicates breakpoints 
				      :test #'= 
				      :key #'car) 
		   #'< :key #'car))))

(defun env-concatenate (&rest envs) 
  (if (= (length envs) 1)
      (copy-list (first envs))
    (let ((xoff 0)
	  (ne nil))
      (loop for env in envs do
	(incf xoff (- .01 (first env)))
	(loop for x in env by #'cddr and y in (cdr env) by #'cddr do
	  (push (+ x xoff) ne)
	  (push y ne))
	(incf xoff (second ne)))
      (nreverse ne))))

(defun env-funcall (function envelope)
  (loop for x in envelope by #'cddr and y in (cdr envelope) by #'cddr
    collect (funcall function x y)))

(defun env-apply (function envelope)
  (loop for x in envelope by #'cddr and y in (cdr envelope) by #'cddr
    collect (apply function (list x y))))

(defun env-map (function envelope &rest more-envelopes)
  (let ((first-group (env-funcall function envelope))
	(second-group (loop for i in more-envelopes 
		       append (env-funcall function i))))
    (apply #'merge-breakpoints (append first-group second-group))))


;;; what we actually want here is to run through all the envelopes in parallel (as per mapcar, I think)
;;; using list-interp to get y at the current x and then call the function on that list of y's:

(defun funcall-breakpoint (function x y)
  (list x (funcall function x y)))

(defun map-across-envelopes (function &rest envelopes)
  (let ((new-e nil)
	(current-x (loop for x in envelopes minimize (car x)))
	(last-x (loop for x in envelopes maximize (env-last-x x))))
    (setf new-e 
      (funcall-breakpoint function 
			  current-x 
			  (loop for x in envelopes 
			   collect (list-interp current-x x))))
    (loop until (>= current-x last-x) do
      (setf current-x (loop for x in envelopes 
		       minimize (or (loop for y in x by #'cddr 
				     if (> y current-x) 
				     return y) 
				    last-x)))
      (setf new-e (append new-e 
			  (funcall-breakpoint function 
					      current-x 
					      (loop for x in envelopes 
					       collect (list-interp current-x x))))))
    new-e))
      
(defun env+ (&rest envelopes) 
  (apply #'map-across-envelopes 
	 (append (list #'(lambda (x y) 
			   (declare (ignore x)) 
			   (apply #'+ y))) 
		 envelopes)))

(defun env* (&rest envelopes) 
  (apply #'map-across-envelopes 
	 (append (list #'(lambda (x y) 
			   (declare (ignore x)) 
			   (apply #'* y))) 
		 envelopes)))

(defun env-max (&rest envelopes) 
  (apply #'map-across-envelopes 
	 (append (list #'(lambda (x y) 
			   (declare (ignore x)) 
			   (apply #'max y))) 
		 envelopes)))

#|
;;; and so on -- here are env/ and env- 
(defun env/ (&rest envelopes) 
  (apply #'map-across-envelopes 
	 (append (list #'(lambda (x y) 
			   (declare (ignore x)) 
			   (apply #'/ y))) 
		 envelopes)))

(defun env- (&rest envelopes) 
  (apply #'map-across-envelopes 
	 (append (list #'(lambda (x y) 
			   (declare (ignore x)) 
			   (apply #'- y))) 
		 envelopes)))
|#


(defun add-or-edit-breakpoint (fn x y)
  (if (null fn)
      (list x y)
    (let ((lim (length fn)))
      (loop for px in fn by #'cddr and
                i from 0 by 2 do
	(if (= px x)
	    (if (= i 0)
		(return-from add-or-edit-breakpoint (append (list x y) (subseq fn 2 lim)))
	      (if (>= (+ i 2) lim)
		  (return-from add-or-edit-breakpoint (append (subseq fn 0 i) (list x y)))
		(return-from add-or-edit-breakpoint (append (subseq fn 0 i) (list x y) (subseq fn (+ i 2) lim)))))
	  (if (> px x)
	      (if (= i 0)
		  (return-from add-or-edit-breakpoint (append (list x y) fn))
		(return-from add-or-edit-breakpoint (append (subseq fn 0 i) (list x y) (subseq fn i lim)))))))
      (append fn (list x y)))))

(defun remove-breakpoint (fn x)
  (if fn
      (let ((lim (length fn)))
	(if (= (first fn) x)
	    (cddr fn)
	  (if (= (env-last-x fn) x)
	      (subseq fn 0 (- lim 2))
	    (if (> lim 2)
		(progn
		  (loop for x0 in (cddr fn) by #'cddr and i from 2 by 2 do
		    (if (= x0 x)
			(return-from remove-breakpoint (append (subseq fn 0 i) (subseq fn (+ i 2) lim)))))
		  (error "cannot find breakpoint at ~A in ~A" x fn))))))))

(defun point-on-line-p (px py qx qy tx ty)

  ;; is point tx ty on line defined by px py and qx qy --
  ;; nil if no, :before if on ray from p, :after if on ray from q, :within if between p and q
  ;; (these are looking at the "line" as a fat vector drawn on a grid)
  ;; taken from "Graphics Gems" by Glassner, code by A Paeth

  (if (or (= py qy ty) (= px qx tx))
      :within
    (if (< (abs (- (* (- qy py) (- tx px))
		   (* (- ty py) (- qx px))))
	   (max (abs (- qx px))
		(abs (- qy py))))
	(if (or (and (< qx px) (< px tx))
		(and (< qy py) (< py ty)))
	    :before
	  (if (or (and (< tx px) (< px qx))
		  (and (< ty py) (< py qy)))
	      :before
	    (if (or (and (< px qx) (< qx tx))
		    (and (< py qy) (< qy ty)))
		:after
	      (if (or (and (< tx qx) (< qx px))
		      (and (< ty qy) (< qy py)))
		  :after
		:within)))))))

(defun env-simplify (env &optional (ygrid 10) (xgrid 100))

  ;; grid = how fine a fluctuation we will allow.
  ;; the smaller the grid, the less likely a given bump will get through
  ;; original x and y values are not changed, just sometimes omitted.

  (if (and env
	   (> (env-length env) 2))
      (let* ((new-env (list (second env) (first env)))
	     (ymax (loop for y in (cdr env) by #'cddr maximize y))
	     (ymin (loop for y in (cdr env) by #'cddr minimize y))
	     (xmax (env-last-x env))
	     (xmin (first env)))
	(if (= ymin ymax)
	    (list xmin ymin xmax ymax)
	  (let* ((y-scl (/ ygrid (- ymax ymin)))
		 (x-scl (/ (or xgrid ygrid) (- xmax xmin)))
		 (px nil) (py nil)
		 (qx nil) (qy nil) 
		 (tx nil) (ty nil) 
		 (qtx nil) (qty nil))
	    (loop for ttx in env by #'cddr and
	              tty in (cdr env) by #'cddr do
	      (setf tx (round (* ttx x-scl)))
	      (setf ty (round (* tty y-scl)))
	      (if px
		  (if (not (point-on-line-p px py qx qy tx ty))
		      (progn
			(push qtx new-env)
			(push qty new-env)
			(setf px qx
			      py qy)))
		(setf px qx
		      py qy))
	      (setf qx tx
		    qy ty
		    qtx ttx
		    qty tty))
	    (push qtx new-env)
	    (push qty new-env)
	    (nreverse new-env))))
    (copy-seq env)))
	       


(defun reduce-amplitude-quantization-noise (e dur amp &optional (ramp-dur .5) (low-amp .005))
  ;; at very low amplitude (0.0 to .005), a ramp can cause an irritating sweeping buzz.
  (let ((new-e nil))
    (when (and e
	       (> (length e) 2)
	       (> dur ramp-dur)
	       (plusp amp))
      (let* ((last-x (env-last-x e))
	     (first-x (car e))
	     ;; for now assume e is 0..1 on y axis
	     (x-dur (* ramp-dur (/ (- last-x first-x) dur)))
	     (y-val (/ low-amp amp))
	     (x0 first-x)
	     (y0 (cadr e)))
	(loop for x1 in (cddr e) by #'cddr and y1 in (cdddr e) by #'cddr do
	  (if (and (> (- x1 x0) x-dur)	;segment is longer than ramp-dur
		   (/= y0 y1))		;only ramp gives this buzz (otherwise "normal" noise)
	      (if (< y0 y-val)		;left value is below low-amp
		  (if (< y0 y1)		;going up => make it go up sooner
		      (setf new-e (add-or-edit-breakpoint (or new-e (copy-seq e))
							  (+ x0 x-dur) (min y1 (max y-val (eref e (+ x0 x-dur))))))
					;going down => delay down-slope as long as possible
		    (setf new-e (add-or-edit-breakpoint (or new-e (copy-seq e))
							  (- x1 x-dur) (min y0 (max y-val (eref e (- x1 x-dur)))))))
		(if (< y1 y-val)	;y0>y-val here, so we are going down and need not check for both y0<y-val and y1<y-val
		    (setf new-e (add-or-edit-breakpoint (or new-e (copy-seq e))
							(- x1 x-dur) (max y-val (eref e (- x1 x-dur))))))))
	  (setf x0 x1)
	  (setf y0 y1))))
    (or new-e e)))


(defun meld-envelopes (e0 e1)
  (if (not e0) 
      e1
    (if (not e1)
	e0
      (let* ((e0-last-x (env-last-x e0))
	     (x-scl (/ 1.0 e0-last-x)))
	(map-across-envelopes #'(lambda (x y)
				  (if (>= x e0-last-x)
				      (cadr y)
				    (+ (car y) (* (- (cadr y) (car y))
						  (* x x-scl)))))
			      e0 e1)))))


;;; fft-env-simplify uses fft filtering to smooth an envelope -- aimed at Sansy envelopes -- there
;;; are lots of variations on this idea depending on the application.  In "Numerical Recipes" they
;;; remove the linear trend of the data, but I have found that causes confusion when the underlying
;;; envelope is really a randomly jittered sine wave without any true linear trend.  Also, here I
;;; believe the min and max values and rescale the result to match the ambitus of the input -- 
;;; cave canem or whatever.

(defun fft-env-simplify (e &optional cutoff)
  (let* ((min-step (loop for x0 in e by #'cddr and 
		             x1 in (cddr e) by #'cddr 
		    minimize (- x1 x0)))
	 (x-extent (- (env-last-x e) (first e)))
	 (x0 (first e))
	 (pts (floor (/ x-extent min-step)))
	 (n (expt 2 (ceiling (log (1+ pts) 2.0))))
	 (cut-n (floor (* n (or cutoff .5))))
	 (datar (make-array n :element-type 'float :initial-element 0.0))
	 (datai (make-array n :element-type 'float :initial-element 0.0)))
    (loop for x from x0 by min-step and i from 0 to pts do
      (setf (aref datar i) (list-interp x e)))
    (let* ((min-y0 (loop for i from 0 to pts minimize (aref datar i)))
	   (max-y0 (loop for i from 0 to pts maximize (aref datar i)))
	   (midpt (* .5 (+ min-y0 max-y0))))
      (loop for i from 0 to pts do
	(decf (aref datar i) midpt))
      (_fft datar datai n 1)    
      (loop for i from cut-n below n do
	(setf (aref datar i) 0.0)
	(setf (aref datai i) 0.0))
      (_fft datar datai n -1)
      (let* ((min-y1 (loop for i from 0 to pts minimize (aref datar i)))
	     (max-y1 (loop for i from 0 to pts maximize (aref datar i)))
	     (inv-n (/ (- max-y0 min-y0) (- max-y1 min-y1)))
	     (y-off (- (* inv-n min-y1) min-y0)))
	(loop for i from 0 to pts and x from 0.0 by min-step 
	 collect x 
	 collect (- (* inv-n (aref datar i)) y-off))))))



(defun lastx (env)
  (nth (- (length env) 2) env))

(defun env-there-and-back-again (env)
  (let* ((x-max (lastx env))
	 (new-env (reverse env))
	 (rev-env (reverse env)))
    (loop for x in (cdddr rev-env) by #'cddr and y in (cddr rev-env) by #'cddr do
      (push (+ x-max (- x-max x)) new-env)
      (push y new-env))
    (nreverse new-env)))

(defun x-norm (env xmax)
  ;; change x axis values so that they run to xmax
  (let ((scl (/ xmax (lastx env))))
    (loop for x in env by #'cddr and y in (cdr env) by #'cddr 
      collect (* x scl) collect y)))

;;; repeat-env will repeat an envelope the number of times specified by its second argument.
;;; Hence if you specify (repeat-env '(0 0 100 1) 2) the result will be (0 0 100 1 101 0 201 1).
;;; Because the final y value was different from the first y value, a quick ramp will be 
;;; inserted between repeats.  You can have every other repeat be a reflection of the given 
;;; envelope by setting the optional second argument to t.  In that case,
;;; (repeat-env '(0 0 100 1) 2 t) the result will be (0 0 100 1 200 0).  If 
;;; you want the original x axis limits respected by the resultant envelope, set
;;; the third argument ("x-normalized") to t.

(defun env-repeat (ur-env ur-num-times &optional reflected x-normalized)
  (let* ((env (if reflected (env-there-and-back-again ur-env) ur-env))
	 (num-times (if reflected (floor ur-num-times 2) ur-num-times))
	 (x-max (lastx env))
	 (first-y (second env))
	 (x (first env))
	 (first-y-is-last-y (= first-y (first (last env))))
	 (offset (/ x-max 100.0))
	 (dxs (loop for x1 in (cddr env) by #'cddr and x0 in env by #'cddr collect (- x1 x0)))
	 (new-env (list first-y x))
	 (repeating nil))
    (loop for i from 1 to num-times do
      (when (and repeating (not first-y-is-last-y))
	(push (incf x offset) new-env)
	(push first-y new-env))
      (loop for dx in dxs and y in (cdddr env) by #'cddr do
	(push (incf x dx) new-env)
	(push y new-env))
      (setf repeating t))
    (when (and reflected (oddp ur-num-times))
      (loop for xx in (cddr ur-env) by #'cddr and dx in dxs and y in (cdddr env) by #'cddr do
	(push (incf x dx) new-env)
	(push y new-env)))
    (if x-normalized
	(x-norm (nreverse new-env) (lastx ur-env))
      (nreverse new-env))))


;;; Here's another useful function, courtesy Michael Edwards (for use with src-change and such)

(defun gliss-dur (env initial-transp input-dur &key (scaler 1) (semitones nil))
  (flet ((semitone (n) (expt 2.0 (/ n 12.0))))
    (let* ((x-max (lastx env))
	   (src (if semitones (semitone initial-transp) initial-transp))
	   (scaled-env (loop for x in env by #'cddr and y in (cdr env) by #'cddr
			  collect x
                          collect (if semitones 
				      (- (semitone (* scaler y)) 1)
				    (* scaler y))))
	   (mean-y (loop for x1 in scaled-env by #'cddr
			 and y1 in (cdr scaled-env) by #'cddr
			 and x2 in (cddr scaled-env) by #'cddr
			 and y2 in (cdddr scaled-env) by #'cddr
			 sum (* (/ (- x2 x1) x-max)
				(+ (min y1 y2) (/ (- (max y1 y2) (min y1 y2)) 2.0))))))
      (/ input-dur (+ src mean-y)))))
