;;; **********************************************************************
;;; Copyright (c) 89-93, 94 Heinrich Taube.  All rights reserved.
;;; Use and copying of this software and preparation of derivative works
;;; based upon this software are permitted and may be copied as long as 
;;; no fees or compensation are charged for use, copying, or accessing
;;; this software and all copies of this software include this copyright
;;; notice.  Suggestions, comments and bug reports are welcome.  Please 
;;; address email to: hkt@zkm.de
;;; **********************************************************************

;;;
;;; This file implements the algorithmic interface to object creation.
;;; Constructor macro are defined for each type of object. These
;;; macros all expand as calls to make-object.
;;;

(in-package :stella)

;;;
;;; make-object returns a new or redefined object. spec is either the name
;;; of a class of object to create, or a list '(class name), where if name
;;; is an existing object it is destructively redefined otherwise a new
;;; object is created with that name.
;;;

(defvar *container* nil)   ; lambda arg in thread and merge expansions.
(defvar *element* nil)     ; lambda arg in algorithm expansion.
(defvar *let-bindings*)    ; lexical vars used in algorithm function.
(defvar *algorithm*)       ; lambda arg in algorithm expansion.

(defun make-object (spec &rest inits)
  (let (class id? id object)
    (if (consp spec)
        (progn 
          (setf id? (consp (cdr spec))) 
          (setf class (find-class (car spec)) id (cadr spec))
          (when id (setf object (find-object id nil))))
      (setf class (find-class spec)))
    (if object
        (progn (unless (eq class (class-of object))
                 (change-class object class))
               (apply #'redefine-object object :id id inits))
      (if id?			; id could be explicitly nil
          (setf object 
            (let ((new (apply #'make-instance class :id id inits))) 
              (when (and id (not *merge*))
                (update-top-level :add new))
              new))
        (setf object (apply #'make-instance class inits))))
    (when *container* (add-object object *container*))
    object))

;;;
;;; redefine object "remakes" an object, ie after the redefinition the
;;; object should behave as if it were actually a new instance, but it
;;; is still eq to its old form.
;;;

(defparameter *warn-if-redefine-object* t)

(defmethod redefine-object ((instance standard-object) &rest inits)
  (when *warn-if-redefine-object* (warn "Redefining ~S." instance))
  (object-makunbound instance)
  (apply #'initialize-instance instance inits)
  instance)
  
(defmethod redefine-object :before ((object container) &rest initargs)
  (declare (ignore initargs))
  (remove-all-objects object))

;;; when we redefine an object we still want it to contain the same
;;; backpointers to its containers as it had before the redefinition.

(defmethod redefine-object :around ((object container-mixin) &rest initargs)
  #-aclpc (declare (ignore initargs))
  (let ((container (slot-value-or-default object 'container))
        (redefined (call-next-method)))
    (when container
	  (setf (slot-value redefined 'container) container))
    redefined))	  

;;;
;;; object is the constructor macro for musical data.
;;;

(defmacro object (type &rest initargs)
  (let ((class (find-class type nil)))
    #+aclpc (unless (class-finalized-p class) (finalize-inheritance class))
    (unless (and class (typep (class-prototype class) 'element))
      (error "~A is not an element class." type))
    `(make-object ',type ., (expand-initargs class initargs))))  

;;;
;;;
;;;
  
#|
(defun expand-container (type name inits body)
  (unless (consp name) (setf name `(quote ,name)))
  `((lambda (*container*) ,@body *container*)
    (make-object (list ',type ,name)
	            .,(expand-initargs (find-class type) inits))))
|#

(defun expand-container (type name inits body)
  (unless (consp name) (setf name `(quote ,name)))
  `((lambda (*container*) 
      ,.(loop for form in body 
              collect (if (and (listp form) (eq (car form) 'find-object))
                          `(add-object ,form *container*)
                        form))
      *container*)
    (make-object (list ',type ,name)
	            .,(expand-initargs (find-class type) inits))))

  
(defmacro thread (name inits &body body &environment environment)
  (declare (ignore environment))
  (expand-container 'thread name inits body)) 

(defmacro heap (name inits &body body &environment environment)
  (declare (ignore environment))
  (expand-container 'heap name inits body)) 

(defmacro merge (name inits &body body &environment environment)
  (declare (ignore environment))
  (expand-container 'merge name inits body))

;;;
;;; algorithm, mute and generator
;;;

(defmacro algorithm (&whole whole name note-class inits &body body
                     &environment environment)
  (expand-algorithm 'algorithm name note-class inits body whole environment))

(defmacro mute (&whole whole name inits &body body
                &environment environment)
  (expand-algorithm 'algorithm name nil inits body whole environment))

(defmacro generator (&whole whole name note-class inits &body body
                     &environment environment)
  (expand-algorithm 'generator name note-class inits body whole environment))

;;;
;;; algorithm expansion rewrites the body of the algorithm as a lisp
;;; function with two args, *algorithm* and *element*. This function is 
;;; funcalled and passed the current algorithm and data objects by the 
;;; scheduler. The lambda args are declared as special variables so that 
;;; they may also be referenced by lisp code outside the lexical scope 
;;; of the algorithm's definition.
;;; Slot names appearing as variable references in the body of the algorithm 
;;; or in the vars statement are rewritten to slot-value forms. Item stream
;;; definitions appearing in the body are rewritten as lexical variable
;;;  bindings.
;;;

(defun add-binding (var &optional value)
  (push (list var value) *let-bindings*)
  var)

;;;
;;; any call to the item function inside the body of an algorithm is wrapped
;;; inside a form that (possibly) updates or side effects the current status
;;; of the running algorithm.
;;;

(defun wrap-item (stream kill? chords? rests?)
  (let ((v1 (gentemp "V"))
        (v2 (gentemp "V"))
        (real-status '(algorithm-status *algorithm*))
        (status (gentemp "S"))
        body)

    ;; since :end-of-period and :chording are disjoint states we try to
    ;; combine the checks into a single IF statement. If kill?
    ;; is an integer the ending check includes the additional check for
    ;; however many periods has been indicated.
    (when kill?
      (let ((end-form `(setf ,status (logior ,status +ending+))))
        (when (and (numberp kill?) (> kill? 1))
          (let ((counter (gentemp "K")))
            (add-binding counter 1)
            (setf end-form `(and (> (setq ,counter (1+ ,counter)) 
                                    ,(coerce kill? 'integer))
                                 ,end-form))))
        (unless (eq kill? t)
          (when (symbolp kill?)
            (let ((counter (gentemp "K")))
              (add-binding counter 1)
              (setf end-form `(and (> (setq ,counter (1+ ,counter)) 
                                   ,kill?)
                                   ,end-form)))))
        (setf body `((if (eq ,v2 +eop+) ,end-form)))))

    (when chords?
      (setf body `((if (eq ,v2 ':chording)
		       (setf ,status (logior ,status +chording+))
		       ,.body))))

    ;; resting and end-of-period are compatible states so we must keep
    ;; the status checks separate.
    (when rests? 
      (setf body `((if (restp ,v1) 
                       (setf ,status (logior ,status +resting+))) 
                     ,.body)))

    ;; if we have rewritten slot accessors, we need to initialize the
    ;; local variable with the actual status form and then do a setf of
    ;; that form at the end.
    ;; we only make a wrapper if we are actually doing status checks.
    (if body
        (progn
          (setf body `((let ((,status ,real-status))
                         ,.body
                         (setf ,real-status ,status))))
          (walk-form `(multiple-value-bind (,v1 ,v2) (item ,stream)
                         ,.body
                         (values ,v1 ,v2))))
      `(item ,stream))))

(defun walk-item (form environment)
  (apply 
    #'(lambda (stream &key kill (rest t) (chord t))
        (declare (special terminates?))
        (when kill
          (setf terminates? t)
          (when (numberp kill)
            (unless (and (integerp kill) (> kill 0))
              (error ":KILL ~S not greater than 0."
                     kill))))
        (when (and (consp stream) 
                   (symbolp (car stream))
                   (or (eq (car stream) 'make-instance)
                       (get (car stream) :item-expand)))
          (let ((var (gentemp "I")))
            (add-binding var (macroexpand stream environment))
            (setf stream var)))
        (values (wrap-item stream kill chord rest)
                t))
    (cdr form)))

;(defun walk-algorithm (form context environment)
;  (if (and (listp form) (eq (car form) 'item)
;           (eq context :eval))
;      (walk-item form environment)
;    (values form nil)))

(defun walk-algorithm (form context environment)
  (declare (special accessors)) ; bound in expand-algorithm
  (cond ((listp form)
         (cond ((and (eq (car form) 'item) (eq context :eval))
                (walk-item form environment))
               (t
                (values form nil))))
        ((symbolp form)
         (when (and (find form accessors :key #'car)
                    (or (walker:variable-lexical-p form environment)
                        (walker:variable-special-p form environment)))
           (warn "Slot ~S will be shadowed by a variable named ~S."
                 form form))         
          
         (values form t))
        (t
         (values form nil))))

(defun walk-vars (statement accessors)
  (dolist (spec (cdr statement)) ; car = VARS
    (if (consp spec)
        (let ((old spec)
              (var (pop spec)) 
              (val (pop spec)))
          (unless (and (symbolp var) (not (constantp var)))
            (error "Vars declaration: ~S not a variable name or binding."
                    var))
          (when spec
            (error "Vars declaration: ~S not a variable binding." old))
          (add-binding var
                       (walk-form `(symbol-macrolet ,accessors ,val)
                                   nil #'(lambda (f c e) c e f))))
      (if (and (symbolp spec) (not (constantp spec)))
          (add-binding spec)
          (error "Vars declaration: ~S not a variable name or binding."
                 spec)))))

;;;
;;; expand-algorithm is the macroexpansions function for the algorithm
;;; and generator macros. It expands to a call to make-object with the
;;; body of the algorithm rewritten to a lisp function.
;;;

(defparameter *break-if-no-obvious-termination*	nil)

(defun expand-algorithm (type name note-class inits body whole environment)
  (unless note-class (setf note-class 'rhythmic-element))
  (let ((saved inits)
        (iclass (find-class note-class nil))
        (gclass (find-class type))
        (terminates? nil)
        (*let-bindings* '())
        ginitargs iinitargs form accessors fun)
    (declare (special terminates? accessors))
    #+aclpc (unless (class-finalized-p iclass) (finalize-inheritance iclass))
    #+aclpc (unless (class-finalized-p gclass) (finalize-inheritance gclass))
    (unless (and iclass (typep (class-prototype iclass) 'element))
      (error "~S is not a legal element class." note-class))
    (loop with gslots = (class-slots gclass)
          and  islots = (class-slots iclass)
          and arg and init and value
          while inits
          do (setf init (pop inits))
             (setf arg (if (constantp init) init `(quote ,init)))
             (if inits 
                 (setf value (pop inits))
               (error "Uneven initialization list: ~S" saved))
             (unless terminates?
               (setf terminates? 
                 (and (member init '(length events end :length :events :end) 
                              :test #'eq)
                      t)))
          if (find init gslots :key #'slot-definition-initargs :test #'member)
          nconc (list arg value) into ginits
          else
          if (find init islots :key #'slot-definition-initargs :test #'member)    
          nconc (list arg value) into iinits
          else 
          do (error "~S is not a legal initialization argument for ~A or ~A."
                        init (class-name iclass) (class-name gclass))
          finally (setf ginitargs ginits iinitargs iinits))

    ;; create the macroexpansion list for symbol-macrolet. slot names
    ;; will get rewritten as calls to slot-value that access either the
    ;; algorithm or the data object. the algorithm slots shadow data slots
    ;; with the same name.
    (setf accessors
      (mapcar #'(lambda (s)
                  (let ((n (slot-definition-name s)))
                    (list n `(slot-value *algorithm* ',n)))) 
              (class-slots gclass)))
    (dolist (s (class-slots iclass))
      (let ((n (slot-definition-name s)))
        (unless (assoc n accessors)
          (push (list n `(slot-value *element* ',n))
                accessors))))

    ;; check for VARS declaration, add entries to the binding list.
    (when (and body (consp (car body)) (eq (caar body) 'vars))
      (walk-vars (pop body) accessors))

    ;; walk the body of the algorithm, add stream definitions to the
    ;; bindings list and warn about slots and lisp variables with the 
    ;; same name.
    (setf fun
      (walk-form `(lambda (*algorithm* *element*)
                    (symbol-macrolet ,accessors ,@body))  
                  environment #'walk-algorithm))
    (unless (consp name) (setf name `(quote ,name)))
    (setf form
     `(make-object (list ',type ,name)
                   :class (find-class ', note-class)
                   :instance-init 
                   #'(lambda (*element*)
                       (initialize-instance *element* .,iinitargs)
                       *element*)
                   :function-init
                   #'(lambda (*algorithm* *element*)	
                       ,(if *let-bindings* 
                            `(let* ,(nreverse *let-bindings*)
                               (function ,fun))
                          `(function ,fun)))
                   :external ',whole ., ginitargs))
    (unless terminates? 
      (let ((str (format nil "A nonterminating ~(~A~) is probably being defined.  Since no END, LENGTH or :KILL expression was provided this ~(~A~) will probably produce ~A notes." type type most-positive-fixnum)))
    (if *break-if-no-obvious-termination* 
        (error str)
      (warn str))))
    form))

#|

(pprint (macroexpand '(algorithm x midi-note (length 30 amplitude .5)
  (setf note (item (intervals 0 1 2 3 in random from note)))
  (setf rhythm (pickl .1 .2 .4))
  (setf duration rhythm))))

(pprint (macroexpand '(algorithm x midi-note (length 30 amplitude .5)
  (setf note (item (intervals 0 1 2 3 in random from note)))
  (setf rhythm (pickl .1 .2 .4))
  (setf duration rhythm))))

(let ((start 1))
  (algorithm x midi-note (length 30 amplitude .5)
  (setf note (item (intervals 0 1 2 3 in random from note)))
  (setf rhythm (pickl .1 .2 .4))
  (setf duration rhythm)))

|#

;;;
;;; user utilities check various states of the running algorithm
;;;

(defmacro unless-chording (&body body)
  `(unless (logtest (algorithm-last-status *algorithm*) +chording+)
     ,@body))

(defmacro when-chording (&body body)
  `(when (logtest (algorithm-last-status *algorithm*) +chording+)
     ,@body))

(defmacro unless-ending (&body body)
  `(unless (logtest (algorithm-status *algorithm*) +ending+)
     ,@body))

(defmacro when-ending (&body body)
  `(when (logtest (algorithm-status *algorithm*) +ending+)
     ,@body))

(defmacro unless-resting (&body body)
  `(unless (logtest (algorithm-status *algorithm*) +resting+)
     ,@body))

(defmacro when-resting (&body body)
  `(when (logtest (algorithm-status *algorithm*) +resting+)
     ,@body))

(defun status? (status &optional (object *algorithm*))
  (ecase status
    (:chording (logtest (algorithm-last-status object) +chording+))
    (:resting (logtest (algorithm-status object) +resting+))
    (:ending (logtest (algorithm-status object) +ending+))
    (:killed (logtest (algorithm-status object) +killed+))
    ((:noteon :noteoff :noteupdate)
     (eq (slot-value object 'type) status))
    (:unqueued (logtest (algorithm-status object) +unqueued+))))

;;;
;;; layout
;;;

(defmacro layout (name directive)
  `(parse-layout-directive ',name ',directive))

(defun parse-layout-directive (id ref &optional superior-type num-expected)
  ;; if recursing, superior-type is class we are parsing for, otherwise nil 
  (labels ((parsestart (x)
             (if (numberp x)
               (if (not (< x 0))
                 x
                 (error "negative start time: ~S" x))
               (error "~S is not a start time." x)))
           (parsenum (s p)
             (parsestart (read-from-string (subseq s p)))))
    (cond
     ((listp ref)
      (when num-expected (error "bogus @ in ~S" ref))
      (let* ((typ (ecase (car ref)
                    (mix 'merge)
                    (thread 'thread)
                    (heap 'heap)))
              sec refs)
        (setf refs
          (loop with tail = (cdr ref)
                with last = nil 
                and flag = nil                   ; expect num?
                while tail
                for form = (pop tail)
                for this = (parse-layout-directive nil form typ flag) 
                when (null this)                 ; @
                do (if (eq flag t) (error "bogus @ in ~S" ref)
                     (setf flag t))               
                else when (numberp this)         ; 5
                do (progn 
                     (unless last 
                       (error "start time without object@ in: ~S" ref)) 
                     (setf (cdr last) (parsestart this))
                     (setf flag nil))
                else when (consp this)
                collect
                (progn 
                  (if (eq flag t) (error "bogus @ in ~S" ref) 
                    (setf flag nil))
                  (when (eq (cdr this) :time)    ; foo@
                    (setf (cdr this)
                      (if (numberp (car tail))
                          (parsestart (pop tail))
                        (error "@ without start time in ~S" ref))))
                  this)
                and do (setf last this)))
        (setf sec 
          (make-object (if id (list typ id) typ)
                       :flags (logior +no-link+ +start-unset+)))
        (loop for r in refs
              collect (car r) into subs 
              when (numberp (cdr r)) collect r into starts
              finally (when subs
                        (add-objects subs sec)
                        (when starts 
                          (setf (slot-value sec 'starts) starts))))
        (if (null superior-type) sec (list sec))))
     ((symbolp ref)
      (when num-expected (error "bogus @ in ~S" ref))
      (cond ((eq ref '@)                                ; @
          (unless (eq superior-type 'merge)
               (error "@ outside of mix: ~S" ref))
             nil)
            ((char= (elt (symbol-name ref) 0) #\@)      ; @5
             (unless (eq superior-type 'merge)
               (error "@ outside of mix: ~S" ref))
             (parsenum (subseq (symbol-name ref) 1) 0) )
            ((char= (elt (symbol-name ref)              ; foo@
                         (1- (length (symbol-name ref)))) #\@)
             (let ((s (read-from-string (symbol-name ref) nil nil
                                        :start 0 :end (1- (length ref)))))
               (cons (find-object s) :time)))
            (t
             (let* ((s (symbol-name ref))               ; foo or foo@5
                    (p (position #\@ s)))
               (if p 
                 (if (eq superior-type 'merge)
                   (let ((n (read-from-string s nil nil :end p)))
                     (setf ref (cons (find-object n)
                                     (parsenum s (1+ p)))))
                   (error "@ outside of mix: ~S" ref))
                 (setf ref (list (find-object ref))))
               ref))))
     ((numberp ref) 
      ref)
     (t 
      (error "~S is not a layout directive" ref)))))

